import re

from django import forms
from django.core.exceptions import ValidationError

from .models import Plan, Subscription


class SubscribeForm(forms.Form):
    stripe_token = forms.CharField(max_length=255)
    plan = forms.ModelChoiceField(queryset=Plan.objects.list_published())
    email = forms.EmailField(required=False)
    quiz = forms.CharField(max_length=500)

    def clean_stripe_token(self):
        data = self.cleaned_data['stripe_token']

        if not re.match(r'^(tok|pm)_[0-9a-zA-Z]+$', data):
            raise ValidationError('Token format validation failed')

        return data


class CancelSubscriptionForm(forms.Form):
    subscription_id = forms.CharField(max_length=255)

    def clean_subscription_id(self):
        data = self.cleaned_data['subscription_id']

        if not re.match(r'^sub_[0-9a-zA-Z]+$', data):
            raise ValidationError('Token format validation failed')

        return data



from django.db import models, transaction
from django.conf import settings
from django.contrib.auth.models import User
from django.core import signing
from django.db.models.signals import pre_save
from django.dispatch import receiver

from . import utils

import stripe
stripe.api_key = settings.STRIPE_SECRET_KEY


class PlanManager(models.Manager):
    def list_published(self):
        return self.filter(published=True)


class Plan(models.Model):
    RECURRING_INTERVALS = (
        ('week', 'Week'),
        ('month', 'Month'),
        ('year', 'Year')
    )

    # General
    title = models.CharField(max_length=255)
    published = models.BooleanField(default=True)
    trial_days = models.IntegerField(default=3)

    # Payment details
    price = models.DecimalField(max_digits=10, decimal_places=2)
    recurring_interval = models.CharField(
            choices=RECURRING_INTERVALS,
            max_length=255)
    product_id = models.CharField(max_length=255, null=True, blank=True)
    price_id = models.CharField(max_length=255, null=True, blank=True)

    # Deeplink
    deep_link = models.URLField(max_length=500, null=True, blank=True)

    objects = PlanManager()

    def __str__(self):
        return self.title


class Subscription(models.Model):
    is_active = models.BooleanField(default=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True, null=True, blank=True)

    user = models.ForeignKey(User, on_delete=models.CASCADE)
    plan = models.ForeignKey(Plan, on_delete=models.CASCADE)

    # Stripe details
    customer_id = models.CharField(max_length=255, null=True, blank=True)
    subscription_id = models.CharField(max_length=255, null=True, blank=True)

    # Aditional parameters for deep link
    quiz = models.TextField(null=True, blank=True)
    client_ip = models.GenericIPAddressField(null=True, blank=True)

    def get_token(self):
        """
        Returns Token signed via django.core.signing.Signer.
        :return str:
        """
        signer = signing.Signer()
        return signer.sign_object({
                    'id': self.pk,
                    'sub': self.subscription_id,
                })

    class Meta:
        unique_together = [('user', 'plan')]

    @classmethod
    def create_from_token(cls, token, plan, user, quiz=None, client_ip=None):
        """
        Creates new subscription from given token with plan for given user.
        :param token: Paymnet source (starts with pm_)
                      or payment token (starts with tok_)
        :param plan: Selected plan (Plan object)
        :param user: Subscriber
        :param quiz: URL encoded quiz data (user selected options for depp link)
        :param client_ip: Remote user IP address
        """

        try:
            obj = cls.objects.get(user=user, plan=plan)
        except cls.DoesNotExist:
            obj = cls(user=user, plan=plan)

        with transaction.atomic():
            if not obj.customer_id:
                customer = stripe.Customer.create(
                        email=user.email,
                        description="User #{}".format(user.id))

                obj.customer_id = customer.id

            kwargs = dict(
                    customer=obj.customer_id,
                    items=[
                        {"price": plan.price_id}
                    ],
                    trial_end=utils.get_trial_end(plan.trial_days))

            if token.startswith('tok_'):
                token = stripe.Token.retrieve(token)

                source = stripe.Source.create(
                    type=token.type,
                    token=token.id)

                stripe.Customer.create_source(
                    obj.customer_id,
                    source=source.id)
            elif token.startswith('pm_'):
                pm = stripe.PaymentMethod.retrieve(token)
                stripe.PaymentMethod.attach(pm.id, customer=obj.customer_id)
                kwargs['default_payment_method'] = pm.id

                if pm.billing_details and pm.billing_details.email:
                    user.email = pm.billing_details.email
                    user.save()

                    if customer.email != user.email:
                        stripe.Customer.modify(customer.id, email=user.email)

            sub = stripe.Subscription.create(**kwargs)
            obj.subscription_id = sub.id

            # Update metadata fields
            obj.quiz = quiz
            obj.client_ip = client_ip

            obj.save()

        return obj

    @property
    def deep_link(self):
        """
        Returns application deep link (user quiz data will be appended to plan.deep_link).
        """
        plan = self.plan

        if not plan.deep_link:
            return None

        link = plan.deep_link

        params = list(filter(bool, [
            'subscription_id=%s' % self.subscription_id,
            ('client_ip=%s' % self.client_ip) if self.client_ip else None,
            ('email=%s' % self.user.email) if self.user.email else None
        ]))

        for p in params:
            glue = '&' if '?' in link else '?'
            link = glue.join([link, p])

        return link

    def __str__(self):
        return self.subscription_id


@receiver(pre_save, sender=Plan)
def plan_pre_save(sender, instance, **kwargs):
    if not instance.price_id:
        utils.create_price_on_stripe(instance)
    else:
        utils.update_price_on_stripe(instance)



from django import template

register = template.Library()


@register.filter
def cents_to_usd(cents):
    """
    Returns datetime object for given unix epoch time.
    :param unix_epoch: int Unix epoch time
    :returns str: USD formatted string .02d
    """

    return '%.2f' % (cents / 100.)



# Generated by Django 3.1.1 on 2021-03-19 02:26

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('stripe_subscription', '0002_auto_20210319_0226'),
    ]

    operations = [
        migrations.AddField(
            model_name='plan',
            name='trial_days',
            field=models.IntegerField(default=3),
        ),
    ]



# Generated by Django 3.1.1 on 2021-03-19 02:26

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('stripe_subscription', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='plan',
            name='deep_link',
            field=models.URLField(blank=True, max_length=500, null=True),
        ),
        migrations.AddField(
            model_name='subscription',
            name='quiz',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='plan',
            name='recurring_interval',
            field=models.CharField(choices=[('week', 'Week'), ('month', 'Month'), ('year', 'Year')], max_length=255),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='is_active',
            field=models.BooleanField(default=True),
        ),
        migrations.AlterUniqueTogether(
            name='subscription',
            unique_together={('user', 'plan')},
        ),
    ]



# Generated by Django 3.1.1 on 2020-09-18 01:37

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Plan',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('price', models.DecimalField(decimal_places=2, max_digits=10, unique=True)),
                ('recurring_interval', models.CharField(choices=[('week', 'Week'), ('month', 'Month'), ('year', 'Year')], default='year', max_length=255)),
                ('product_id', models.CharField(blank=True, max_length=255, null=True)),
                ('price_id', models.CharField(blank=True, max_length=255, null=True)),
                ('published', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True, null=True)),
                ('customer_id', models.CharField(blank=True, max_length=255, null=True)),
                ('subscription_id', models.CharField(blank=True, max_length=255, null=True)),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='stripe_subscription.plan')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ]
        ),
    ]

    perations = [
        migrations.AlterUniqueTogether(
            name='Subscription',
            unique_together={('user', 'plan')},
        ),
    ]
